-- STANDART LUA SHELL FOR TTS SOFTWARE "DEMAGOG"
-- (Automatically loadable module. It must be in ANSI encoding)
-- (c) Benedict Lee

-- Necessary public constants

feUnknown   = 0
feANSI      = 1
feUnocode   = 2
feUnicodeBE = 3
feUTF8      = 4

stNormal    = 1
stNoDup     = 2
stDupOnly   = 3
stUniqOnly  = 4
stInverse   = 5
stLen       = 6
stLenDes    = 7
stNum       = 8
stNumDes    = 9
stDic       = 10

maLua       = 1
maRex       = 2
maDic       = 3

wsHide      = 0
wsNormal    = 1
wsMinimize  = 2

btEsc       = 0
btOK        = 1


-- Important functions for working with UTF-8 encoding
-- Loading text file into string
-- file encoding can be ANSI or Unicode with BOM, Result - ANSI-string
-- Note: UTF-8 without BOM will be loaded as encoded ANSI-string
function LoadText(fname)
  return SuperRepl(fname,nil,nil,9)
end


function Utf8ToAnsi(s)
  local bom = '\239\187\191'
  if string.sub(s,1,3) == bom then s = string.sub(s,4,#s) end
  return SuperRepl(s,nil,nil,10)
end


function AnsiToUtf8(s, hasbom)
  local bom = '\239\187\191'
  local r = SuperRepl(s,nil,nil,11)
  if hasbom then r = bom..r end
  return r
end


function DosToAnsi(s)
    return OtherUtils(s,251)
end
DosToWin = DosToAnsi


function AnsiToDos(s)
    return OtherUtils(s,252)
end
WinToDos = AnsiToDos


-- The most important function of dictionary replacements in the text
-- Text from tab i
-- Result to tab j
-- d - dictionary name or name of the dictionaries names table
-- g = true/false - display or not display the progress bar
function WFilter(i, j, d, g)
  local x
  local c
  if g == nil then g = true end
  if type(d) == 'table' then
    -- get the textual representation of the dictionary table
    x = ''
    for k, v in pairs(d) do
      c = k..'='..table.concat(v,'|')
      if x == '' then
        x = c
      else
        x = x..'||'..c
      end
    end
  else
    x = d
  end
  WSuperFilter(i, j, x, g)
end


-- Rule statistics
-- Text from tab i
-- Modified text to tab j
-- List of rules in the tab k
-- (if j = k, then both results are combined in one tab)
-- d - dictionary name or name of the dictionaries names table
-- g = true/false - display or not display the progress bar, default true
function WStat(i, j, k, d, g)
  local x
  local c
  if g == nil then g = true end
  if type(d) == 'table' then
    -- get the textual representation of the dictionary table
    x = ''
    for k, v in pairs(d) do
      c = k..'='..table.concat(v,'|')
      if x == '' then
        x = c
      else
        x = x..'||'..c
      end
    end
  else
    x = d
  end
  WSuperStat(i, j, k, x, g)
end


-- The same WFilter(), but with automatic algorithm selection
-- application of dictionaries, based on the presence in the file name
-- of the characters y, z ( recommended characters: y = '@', z = '^' )
-- y - yes / no - Direct Enumeration / Fast Algorithm
-- z - yes / no - sorting rules in descending of length / without sorting
-- Thank tonio_k
function WFilterAutoDicAlgorithm(i, j, d, g, y, z)
    local o = {}
    -- default - F.A. without sorting rules
    o.Reading_HashAlgorithm = true; o.Reading_OrderedRules = false
    if y ~= nil and #y > 0 then
       o.Reading_HashAlgorithm = not string.match(d,y)
    end
    if z ~= nil and #z > 0 then
       o.Reading_OrderedRules = string.match(d,z)
    end
    Settings(o)  -- store the settings!
    WFilter(i, j, d, g)
end  


-- Redefinition Lua-function 'print', because in Lua: TVerySimpleLua I defined
-- the property Lua.OnPrint without linebreak on end! (it's convenient for me)

put = print

function newline()
  put('\r\n')
end

function print(...)
  put(...)
  newline()
end


-- Exponentially distributed pseudo-random numbers with average value of m
function randexp(m)
  return -m*math.log(1-math.random())
end


-- Gamma distributed pseudo-random numbers with range k and average value of m
function randgamma(k,m)
  local r = 1
  for i = 1, k do
    r = r*(1-math.random())
  end
  return -m*math.log(r)/k
end


-- Normal distributed pseudo-random numbers with average of m and deviation st
function randnormal(m,st)
  local sum = 0
  local r
  for i = 1, 12 do
    r = math.random()
    sum = sum+r
  end
  return st*(sum-6)+m
end


-- Pseudo random number in (0,1) for integer i in [1,2147483647]
function irontab(i)
  local f, maxint = 0.618033988749894848, 2147483647
  if i < -maxint then i = -i+maxint end
  if i > maxint then i = i-maxint end
  return f*i % 1
end


-- Simple hash function for strings
-- Generates numeric hashes from 0 to n-1
-- (it is strongly recommended that n be a prime number)
function  Hash(s, n)
    if not n then n = 1000003 end
    M = 31
    R = 0
    for  i = 1, #s do
      R = (M*R+string.byte(s,i)) % n
    end
    return R
end


-- There is no standart "round" function in Lua, so we make it
function round(num,idp)
  local mult = 10^(idp or 0)
  return math.floor(num*mult+0.5)/mult
end


-- Table of lines without duplicates
-- Thank tonio_k
function table.unique(a)
    local q, t = {}, {}
        for  i = 1, #a do
            if not q[a[i]] then
                 q[a[i]] = true
                 t[#t+1] = a[i]       
            end
    end
    return t
end


-- Table of lines duplicates only
-- dupall = true - duplicates together with the originals
-- Thank tonio_k
function table.duplicates(a, dupall)
    local q, t, d = {}, {}, {}
    for  i = 1, #a do
        if not q[a[i]] then
            q[a[i]] = true
        else
            if dupall and not d[a[i]] then
               d[a[i]] = true
               t[#t+1] = a[i]
               t[#t+1] = a[i]
            else
               t[#t+1] = a[i]
            end
        end
    end
    return t
end


-- Sorting table of lines without duplicates
function table.sortuniq(lines)
    lines = table.unique(lines)
    table.sort(lines)
    return lines
end


-- Save table to file
function table.save(tbl,file)
        -- recursively saves table data to a file, human-readable and indented
        -- supported types: number, string, boolean, tables of the same
        local f,err = io.open(file,"w")
        if err then
                print("Error: Unable to open file for writing: "..file.."\nDetails: "..err)
                return
        end
        local indent = 1
        --
        -- local functions to make things easier
        local function exportstring(s)
                s=string.format("%q",s)
                s=s:gsub("\\\n","\\n")
                s=s:gsub("\r","")
                s=s:gsub(string.char(26),""..string.char(26).."")
                return s
        end
        local function serialize(o)
                if type(o) == "number" then
                        f:write(o)
                elseif type(o) == "boolean" then
                        if o then f:write("true") else f:write("false") end
                elseif type(o) == "string" then
                        f:write(exportstring(o))
                elseif type(o) == "table" then
                        f:write("{\n")
                        indent = indent + 1
                        local tab = ""
                        for i=1,indent do tab = tab .. "        " end
                        for k,v in pairs(o) do
                                f:write(tab .. "[")
                                serialize(k)
                                f:write("] = ")
                                serialize(v)
                                f:write(",\n")
                        end
                        indent = indent - 1
                        tab = ""
                        for i=1,indent do tab = tab .. "        " end
                        f:write(tab .. "}")
                else
                        print("unable to serialzie data: " .. tostring(o))
                        f:write("nil, -- ***ERROR: unsupported data type!***")
                end
        end
        --
        -- here's the actual save process
        f:write("return {\n")
        for k,v in pairs(tbl) do
                f:write("       [")
                serialize(k)
                f:write("] = ")
                serialize(v)
                f:write(",\n")
        end
        f:write("}")
        f:close()
end


-- Load table from file
function table.load(file)
        local data, err = loadfile(file)
        if err then return nil,err else return data() end
end


-- List of all words in string. Dividing produced by the separator character
-- usage example: MyWordList = string.split(MyStr,' ')
function string:split(sep)
  local sep, fields = sep or ":", {}
  local pattern = string.format("([^%s]+)", sep)
  self:gsub(pattern, function(c) fields[#fields+1] = c end)
  return fields
end


-- Simple split string with STRICT delimiters
-- Example: StrSplit( 'qwe|||rty||abc|','|') ==> {'qwe','','','rty','','abc',''}
function StrSplit(s, delim, plain)
    if plain == nil then plain = true end
    if delim == '' then return {} end
    if delim == nil then delim = ' ' end
    local list = {}
    local pos = 1
    while true do
         local first, last = string.find(s, delim, pos, plain)
         if first then -- found?
             list[#list+1] = string.sub(s, pos, first-1)
             pos = last+1
         else
             list[#list+1] = string.sub(s, pos)
             break
         end
    end
    return list
end


function string:trim()
  return (self:gsub("^%s*(.-)%s*$", "%1"))
end


function ValidFileName(fname)
  local i, b
  if fname == nil or fname == '' then return false end
  for i = 1, #fname do
    b = string.byte(fname,i,i)
    if b==34 or b==42 or b==47 or b==58 or b==60 or b==62 or b==63 or b==92 or b==124 then return false end
  end
  return true
end


-- Get file size
function fsize(fname)
  local file = io.open(fname)
  if not file then return -1 end
  local current = file:seek()
  local size = file:seek('end')
  file:seek('set',current)
  file:close()
  return size
end


function FileExists(name)
   local f = io.open(name,"r")
   if f~=nil then io.close(f) return true else return false end
end


--- Check if a file or folder exists in this path
function Exists(f)
   local ok, err, code = os.rename(f,f)
   if ok then res = true else res = false end
   if not res then
      if code == 13 then
         -- Permission denied, but it exists
         res = true
      end
   end
   return res, err
end


--- Check if a directory exists in this path
function FolderExists(path)
   -- "/" works on both Unix and Windows
   return Exists(path.."/")
end
IsFolder = FolderExists


-- List of drive letters
function DriveLetters()
    local alf = 'ABCDEFGHIJKLMNQRSTUVWXYZ'
    local disk = {}
    local d
    for i = 1,#alf do
        d = string.sub(alf,i,i)..':'
        if IsFolder(d..'/') then disk[#disk+1] = d end
    end
    return disk
end


-- Save list of lines to file
function SaveToFile(list,filename)
  local f, err = io.open(filename,"wb")
  if not f then
    return false, err
  end
  for i = 1,#list do
    f:write(list[i]..'\r\n')
    --if i < #list then f:write(list[i]..'\r\n') else f:write(list[i]) end --better?
  end
  f:close()
  return true, nil
end


-- Load list of lines from file
-- This version use the functions from project "Demagog"
-- From the file in any encoding will extract ANSI-string!
function LoadFromFile(fname,allformats)
  if allformats then  -- load file to Demagog Stats-window and convert to Ansi-string
    WOpen(-1,fname)
    local s = WText(-1)
    local L = StrSplit(s,'\r')  -- empty lines also will be included
    WNew(-1)
    return L
  else  -- for Ansi-text files only
    L = {}
    for line in io.lines(fname) do  -- read file in separate lines
      L[#L+1] = line
    end
    return L  -- lines list
  end
end


-- Read file as string of bytes
function StrFromFile(fname)
    local f = io.open(fname, "rb")
    local s = f:read "*a"
    f:close()
    return s
end

-- Save string to file
function StrToFile(s, fname)
    local f = io.open(fname,"w")
    f:write(s)
    f:close()
end


function HomeFolder(s)
  if s == nil then s = '' end
  return OtherUtils(s,0)
end


function CurrentLang()
  return OtherUtils('',1)
end


-- Find voice by short name and return full voice name or nil when voice does not exist
-- If short name is empty or not assugned then return full name of active voice
function Voice(s)
  if s == nil then s = '' end
  local c = OtherUtils(s,2)
  local t = StrSplit(c,'\3')
  if t[1] == '' then t[1] = nil end
  return t[1], t[2], t[3], t[4]
end


function VoiceList()
  local c = OtherUtils('',3)
  local a = string.split(c,'\r\n')
  return a
end


function ShowMessage(c, utf8)
  if utf8 ~= nil then utf8 = '\3utf8\3' end
  MsgDlg(c,utf8,nil,nil,nil,0)
end


function MessageDialog(c)
  return (MsgDlg(c,nil,nil,nil,nil,1) == 1)
end


-- Set voice by short name
function SetVoice(vName, vRate, vPitch, vVolume)
  if vPitch == nil then vPitch = 0 end
  if vRate == nil then vRate = 0 end
  if vVolume == nil then vVolume = 100 end
  MsgDlg(vName..','..vRate..','..vPitch..','..vVolume,nil,nil,nil,nil,5)
end


function StatusMessage(c)
  MsgDlg(c,nil,nil,nil,nil,6)
end


function Sound()
  MsgDlg(nil,nil,nil,nil,nil,7)
end


function Cancel()
  return (MsgDlg(nil,ni,nil,nil,nil,8) == 1)
end


function Settings(o)
    local c = ''
    for key, val in pairs(o) do
        if type(val) == 'boolean' then
            if val then
                c = c..key..'='..'-1'..'\r'
            else
                c = c..key..'='..'0'..'\r'
            end
        end
        if type(val) == 'number' then
            c = c..key..'='..tostring(val)..'\r'
        end
        if type(val) == 'string' then
            c = c..key..'='..val..'\r'
        end
    end
    MsgDlg(c,nil,nil,nil,nil,10)
end


-- DEPRECATED! Not recommended
function FastDic(fastalg, ordered)
    local o = {}
    o.Reading_HashAlgorithm = fastalg
    o.Reading_OrderedRules = ordered
    Settings(o)
end


function ForbidTextAutochange()
    local o = {}
    o.Notes_Auto = false
    o.Eureka = false
    o.Literary_Auto = false
    o.Dictionary_Auto = false
    Settings(o)
end


function Menu(cap, items, ind, width, fontsize)
  local itm = table.concat(items,'\3')
  if ind == nil then ind = 1 end
  if width == nil then width = 686 end
  if fontsize == nil then fontsize = 10 end
  return MsgDlg(cap,itm,ind,width,fontsize,4)
end


function Input(cap, prompts, width, ratio, fontsize)
  -- for compatibility with the old version
  if type(width) == 'boolean' then
    if width then ratio = 0.5 else ratio = 0 end
    width = 686
    fontsize = 10
  end
  --
  if width == nil then width = 686 end
  if ratio == nil then ratio = 0.5 end
  if fontsize == nil then fontsize = 10 end
  local pro = table.concat(prompts,'\3')
  local a = InputBox(cap,pro,width,ratio,fontsize)
  if a == '_NO_ANSWER_' then return nil end
  local b = string.split(a,'\3')
  for i = 1,#b do
    if b[i] == '_EMPTY_' then b[i] = '' end
  end
  return b
end


-- Data Entry User Form --
-- Utility function for Form
function FrmTableToString(d)
    s = ''
    for key, value in pairs(d) do
        if type(value) == 'table' then
            for key2, value2 in pairs(value) do
                if type(value2) == 'boolean' then
                    if value2 then
                        v = 'true'
                    else
                        v = 'false'
                    end
                else
                    if type(value2) == 'number' then
                        v = tostring(value2)
                    else
                        v = value2
                        if key2 == 'Action' then v = v:gsub('\n',' ') end
                    end
                end
                s = s..key..'.'..key2..' = '..v..'\r\n'
            end
        end
    end
    return s
end


function FrmAnswerToTable(s)
    local a = {}
    local b, p, q, r
    a.Answer = btOK
    if s == '' or s == '_NO_ANSWER_' then a.Answer = btEsc; goto FIN end
    s = s:gsub('\n','')
    b = string.split(s,'\r')
    for i = 1,#b do
        p = string.split(b[i],'=')
        if p[1] then
            r = string.split(p[1],'.')
            if r[1] then
                if p[1]:sub(1,4) == 'Edit' or p[1]:sub(1,4) == 'Memo' then
                    if p[2] then a[r[1]] = {['Text'] = p[2]} else a[r[1]] = {['Text'] = ''} end
                end
                if p[1]:sub(1,4) == 'Spin' then
                    if p[2] then a[r[1]] = {['Value'] = p[2]} else a[r[1]] = {['Value'] = 0} end
                end
                if p[1]:sub(1,5) == 'Combo' or p[1]:sub(1,4) == 'List' then
                    if p[2] then
                        q = string.split(p[2],'\3')
                        a[r[1]] = {['Index'] = tonumber(q[1]), ['Item'] = q[2], ['SetText'] = q[3]}
                    end
                end
                if p[1]:sub(1,5) == 'Check' then
                    a[r[1]] = {['Done'] = (p[2] == 'true')}
                end
                if p[1]:sub(1,5) == 'Notes' then
                    if p[2] then
                        q = string.split(p[2],'\3')
                        a[r[1]] = {['Agree'] = q[1], ['AgreeItems'] = q[2]}
                    end
                end
            end
        end
    end
    ::FIN::
    return a
end


-- Displaing the data entry
function Form(layout, upd)
    if type(layout) == 'table' then
        layout = FrmTableToString(layout)
    else
        return {}
    end
    if type(upd) == 'table' then
        upd = FrmTableToString(upd)
    else
        upd = nil
    end
    local s = Data(layout,upd)
    return FrmAnswerToTable(s)
end


function FileAvailable(c)
    return ('1' == OtherUtils(c,249))
end


-- Sets property values Text, SetText, Index, Done
function Let(c)
    if type(c) == 'table' then
        c = FrmTableToString(c)
    else
        c = ''
    end
    OtherUtils(c,5)
end


-- Returns the values of the Text, SetText, Index, Done properties
function Val(c)
    local b = OtherUtils(c,6)
    local k = string.find(c,'.Index',1,true)
    local m = string.find(c,'.Value',1,true)
    local z = string.find(c,'.SetText',1,true)
    if k then 
        b = tonumber(b)
    elseif m then
        b = tonumber(b)
    elseif z then
        b = string.gsub(b,'\n','')
        b = string.gsub(b,'\r','|')
    else
        if b == 'true' then b = true end
        if b == 'false' then b = false end
    end
    return b
end
-- End of Data Entry User Form --


function BeforeShutdown()
    OtherUtils(nil,7)
end


function FileNavigator(cap, path, mask, width, fontsize)
    if fontsize == nil then fontsize = 10 end
    if width == nil then width = 686 end
    if mask == nil then mask = '*.*' end
    local c = cap..'|'..path..'|'..mask..'|'..width..'|'..fontsize
    local b = OtherUtils(c,253)
    if b == '' then b = nil end
    return b
end


function Launch(fname, params, wstate)
    if wstate == nil then wstate = 1 end
    if params == nil then params = '' end
    local c = fname..'|'..params..'|'..wstate
    return ('1' == OtherUtils(c,254))
end


function Execute(fname, params, wstate)
    if wstate == nil then wstate = 1 end
    if params == nil then params = '' end
    local c = fname..'|'..params..'|'..wstate
    return ('1' == OtherUtils(c,255))
end


function Restart(param, safe)
    if param == nil then param = '' end
    if safe then BeforeShutdown() end
    local h = HomeFolder()
    Launch(h..'Demagog.exe',param,wsNormal)
    os.exit()
end


function OpenDialog(multiselect, title, initdir, fltr, fltr_ind)
  local s, c
  if multiselect then c = 'true' else c = 'false' end
  if title == nil then title = '' end
  if initdir == nil then initdir = '' end
  if fltr == nil then fltr = '' end
  if fltr_ind == nil then fltr_ind = -1 end  
  s = c..'|'..title..'|'..initdir..'|'..fltr..'|'..tostring(fltr_ind)
  local a = OtherUtils(s,247)
  if a == '' then return nil end
  if multiselect then
    return string.split(a,'|')
  else
    return a
  end
end


function SaveDialog()
  local d = OtherUtils(nil,248)
  if d == '' then return nil else return d end
end


function AllFiles(folder,mask)
  local a = FilesReview(folder,mask)
  return string.split(a,'|')
end


function AllFolders(folder,mask)
  local a = FilesReview(folder,mask,false)
  return string.split(a,'|')
end


function WAdd(i,j,s)
  if s ~= nil then
    WAppend(i,j,s)
  else
    if type(j) == 'number' then WAppend(i,j,'') else WAppend(i,-2,j) end
  end
end


function WClearSel(i)
   return WSel(i,WSel(i),0)
end


function WSelText(i, c, utf8)
  if utf8 == nil then utf8 = false end
  if c == nil then c = '\3\3\3' end
  if type(c) == 'boolean' and not utf8 then
    utf8 = c
    c = '\3\3\3'
  end
  return WSetSelText(i,c,utf8)
end


-- Add string to 'Stats' tab 
function WLog(s)
    WAppend(0,-2,s)
end


function GaugeInit(max)
  Gauge(-max)
end


-- Fuzzy string comparison
-- The result is a real number from 0 to 1
-- For example, the value 0.5 is quite a good match
function Similar(x, y, k)
  local function Sim(x, y, k)
    local n = #x
    if k > n then k = n end
    local Good = 0
    local Vsego = 0
    local a = x
    local b = y
    local c 
    for i = n-k, n-1 do
      for j = 1, i+1 do
         c = string.sub(a,j,j+n-i-1)
         Vsego = Vsego+1
         if string.find(string.lower(b),string.lower(c),1,true) then Good = Good+1 end
       end
     end
     return Good, Vsego
   end  -- Sim
 -- Provides commutativity
  local Good1, Vsego1, Good2, Vsego2
  if k == nil then k = 3 end
  if k <= 0 then k = 1 end
  Good1, Vsego1 = Sim(x,y,k)
  Good2, Vsego2 = Sim(y,x,k)
  return (Good1+Good2)/(Vsego1+Vsego2)
end


-- Split file name as path, short name, extension
function SplitFileName(s)
    local a = {'','',''}
    local dot = false
    local d = #s+1
    for k = #s,1,-1 do
        b = string.sub(s,k,k)
        if (b == '.') and not dot then
            a[3] = string.sub(s,k,#s)
            d = k 
            dot = true
        end
        if (b == '\\') or (b == '/') then
            a[2] = string.sub(s,k+1,d-1)
            a[1] = string.sub(s,1,k)
            return a  
        end
    end
    a[2] = string.sub(s,1,d-1)
    return a
end


-- Split text into fragments of a given size (in characters) --
--
-- It's approximately. Fragments always a little larger than specified
-- u - template of elementary "brick" of text; for example '[^\r]*[\r]' is a paragraph 
function Fragz(s, fragsize, u)
    local a, c, k, z = {}, '', 1, 1
    if string.sub(s,#s,#s) ~= '\r' then s = s..'\r' end
    while k do
        k, z = string.find(s,u,k)
        if k then
            c = c..string.sub(s,k,z)
            if #c > fragsize then
                 a[#a+1] = c
                 c = ''
            end
            k = z+1
        end
    end
    if c ~= '' then a[#a+1] = c end
    return a
end


-- Substring from i to j with length limitation m
-- v - limitation chars string, for example '.!?\133;,\32'
function LimSub(s, i ,j, m, v)
    local b
    if j-i+1 > m then
        for z = i+m-1,i,-1 do
            b = string.sub(s,z,z)
            if string.find(v,b,1,true) then
                return string.sub(s,i,z), z
            end
        end
        return string.sub(s,i,i+m-1), i+m-1
    else
        return string.sub(s,i,j), j
    end
end


-- Similar to Fragz() but split into fragments NOT LARGE than the given size
function FragzA(s, fragsize, u, v)
    local a, c, k, z = {}, '', 1, 1
    local b
    if string.sub(s,#s,#s) ~= '\r' then s = s..'\r' end
    while k do
        k, z = string.find(s,u,k)
        if k then
            b = string.sub(s,k,z)
            if #c+#b > fragsize then
                if c ~= '' then a[#a+1] = c; c = '' end
                -- limitation too large fragment
                if #b > fragsize then a[#a+1], z = LimSub(s,k,z,fragsize,v); k = z+1 end
            else
                c = c..b
                k = z+1
            end
        end
    end
    if c ~= '' then a[#a+1] = c end
    return a
end


-- Main function of text splitting
function Fragments(s, fragsize, parag)
    if parag == true then return Fragz(s,fragsize,'[^\r]*[\r]') end
    if parag == false then return Fragz(s,fragsize,'[^\r%.!%?\133]*[\r%.!%?\133]') end
    if parag == nil then return FragzA(s,fragsize,'[^\r]*[\r]','.!?\133;,\32') end
end


-- Create a series of small files from big text in tab i
function WSeries(i, folder, fragsize, parag, header, ender, startnum)
    local fname
    local s = WText(i)
    local b = Fragments(s,fragsize-2,parag)
    if string.trim(b[#b]) == '' then b[#b] = nil end
    local k = #folder
    if startnum == nil then
        startnum = 1
    end
    if string.sub(folder,k,k) ~= '/' and string.sub(folder,k,k) ~= '\\' then
        folder = folder..'\\'
    end
    for i = 1,#b do
        fname = folder..string.format("%04d",(i+startnum-1))..'.txt'
        if header ~= nil then
            b[i] = string.format(header,i+startnum-1)..'\r'..b[i]
        end
        if ender ~= nil then
            b[i] = b[i]..'\r'..string.format(ender,i+startnum-1)
        end
        SaveToFile({b[i]},fname)
    end
    return #b
end
-- End of split text... --


-- Apply dictionaries to multiple files in folder
-- diclist = { {'dictionary1', mode'}, {'dictionary2', mode}, ... {'dictionaryN', mode} }
-- mode = true if the dic-dictionary is applied by the fast algorithm and false if it is 
-- brute-force algorithm. 
-- A simplified view of the list of dictionaries is allowed: 
-- diclist = {'dictionary1', ... {'dictionaryX'}, ...} - i.e. in braces is only the name 
-- of the dic-dictionary used by brute-force algorithm 
function FilterToAll(folder, mask, diclist)
    local b, d, mode
    local k = #folder
    if string.sub(folder,k,k) ~= '/' and string.sub(folder,k,k) ~= '\\' then
        folder = folder..'\\'
    end
    local a = AllFiles(folder,mask)
    for i = 1,#a do
        WOpen(0,folder..a[i])
        for j = 1,#diclist do
            b = diclist[j]
            if type(b) == 'table' then
                d = b[1]
                mode = b[2]
            else
               d = b
               mode = true
            end
            FastDic(mode,mode)
            WFilter(0,0,d)
        end
        WSave(0,folder..a[i])
    end
    WNew(0)
    return #a, #diclist
end


-- Create audio for all files in a given folder
function AudioToAll(folder, mask, remsource)
    a = AllFiles(folder,mask)
    for i = 1,#a do
        WOpen(0,folder..a[i])
        WAudio(0,'',false)
        if remsource then os.remove(folder..a[i]) end
    end
    WNew(0)
    return #a
end


-- Simultaneously pply dictionaries and create audio for all files in a given folder 
-- Thank tonio_k
function FilterAndAudioToAll(folder, mask, diclist, remsource)
    ForbidTextAutochange()
    local b, d, p, mode
    local k = #folder
    if string.sub(folder,k,k) ~= '/' and string.sub(folder,k,k) ~= '\\' then
        folder = folder..'\\'
    end
    local a = AllFiles(folder,mask)
    for i = 1,#a do
        WOpen(0,folder..a[i])
        for j = 1,#diclist do
            b = diclist[j]
            if type(b) == 'table' then
                d = b[1]
                mode = b[2]
            else
               d = b
               mode = true
            end
            FastDic(mode,mode)
            WFilter(0,0,d)
        end
        WSave(0,folder..'temp_'..a[i])
        WOpen(0,folder..'temp_'..a[i])
        WAudio(0,folder,false)
        p = SplitFileName(a[i])[2]
        os.rename (folder..'temp_'..p..'.mp3', folder..p..'.mp3')
        os.remove(folder..'temp_'..a[i])
        if remsource and i > 1 then os.remove(folder..a[i-1]) end
    end
    if remsource then os.remove(folder..a[#a]) end
    WNew(0)
    return #a, #diclist
end


-- Remove punctuation at the beginning and end of string
function Body(s) 
  return SuperRepl(s,nil,nil,3)
end


-- Read aloud text from string
function StrSpeak(s)
  WNew(-1,s)
  WSpeak(-1)
end


-- Read aloud text from Clipboard
function CSpeak()
  local s = Clipboard()
  StrSpeak(s)
end


-- Question about the possibility of shutdown the computer
function ShutdownDialog(ind)
    local cap, itms, warn, a
    if CurrentLang() == 'Russian' then
        cap = '    ?'
        itms = {' ',' '}
        warn = '     !'
    else
        cap = 'Shut down computer when script done?'
        itms = {' YES',' NO'}
        warn = 'When script done, the computer will be turned off!'
    end
    if ind == nil then ind = 1 end
    a = Menu(cap,itms,ind)
    if a == 0 then return nil end
    if a == 1 then 
        ShowMessage(warn)
        return true
    end  
    if a == 2  then return false end
end  


-- Thank  balaamster 
function SecondsToClock(seconds)
    local sec = tonumber(seconds)
    local h = math.floor(sec / 3600)
    local m = math.floor((sec % 3600) / 60)
    local s = math.floor(sec % 60)
    local hms = string.format("%02d:%02d:%02d", h, m, s)
    return hms
end


-- Wait for seconds
function pause(sec)
  Grafica(251,sec*1000)
end

sleep = pause  -- synonym for 'pause'


function StrMatch(s, mask)
  local _, k = SuperMatch(s,mask,1,0)
  return (k == -1)
end


function RexMatch(s, mask, spos)
  if spos == nil then spos = 1 end
  local fnd, k = SuperMatch(s,mask,spos,1)
  return fnd, k
end


function DicMatch(s, mask, spos)
  if spos == nil then spos = 1 end
  local fnd, k = SuperMatch(s,mask,spos,2)
  return fnd, k
end


-- s - source text
-- p - regular expression
-- g = true/false (true by default) - display progress bar
function RexRepl(s, p, g)
  local r = table.concat(p,'\r')
  if g == nil then g = true end 
  return SuperRepl(s,r,g,1)
end


-- s - source text
-- p - dic-rule
-- g = true/false (true by default) - display progress bar
function DicRepl(s, p, g)
  local r = table.concat(p,'\r')
  if g == nil then g = true end 
  return SuperRepl(s,r,g,2)
end


-- miscellaneous sort for a list of strings
function Sort(a, st, utf8)
  local s = table.concat(a,'\r')
  if utf8 == nil then utf8 = false end
  if st == nil then st = stNormal end
  s = SuperSort(s,st,utf8)
  a = string.split(s,'\3')
  return a
end


function Markup(s, ru, en)
  return SuperRepl(s,ru..'|'..en,nil,4)
end


function DelMarkup(s)
  return SuperRepl(s,nil,nil,5)
end


function WMarkup(i, j, ru, en)
  local s = WText(i)
  s = SuperRepl(s,ru..'|'..en,nil,4)
  WNew(j)
  WAdd(j,s)
end


function WDelMarkup(i, j)
  local s = WText(i)
  s = SuperRepl(s,'',nil,5)
  WNew(j)
  WAdd(j,s)
end


function RuIPA(s)
  return SuperRepl(s,'',false,6)
end


function RuUPS(s)
  return SuperRepl(s,'',false,7)
end


function IPAtoXS(s)
    local x, b = {}, ''
    --
    x[32] = [[ ]]
    x[865] = [[]]
    x[8255] = [[-\]]
    x[7547] = [[1]]
    --
    x[97] = [[a]]
    x[98] = [[b]]
    x[595] = [[b_<]]
    x[99] = [[c]]
    x[100] = [[d]]
    x[598] = [[d`]]
    x[599] = [[d_<]]
    x[101] = [[e]]
    x[102] = [[f]]
    x[609] = [[g]]
    x[608] = [[g_<]]
    x[104] = [[h]]
    x[614] = [[h\]]
    x[105] = [[i]]
    x[106] = [[j]]
    x[669] = [[j\]]
    x[107] = [[k]]
    x[108] = [[l]]
    x[621] = [[l`]]
    x[634] = [[l\]]
    x[109] = [[m]]
    x[110] = [[n]]
    x[627] = [[n`]]
    x[111] = [[o]]
    x[112] = [[p]]
    x[632] = [[p\]]
    x[113] = [[q]]
    x[114] = [[r]]
    x[637] = [[r`]]
    x[633] = [[r\]]
    x[635] = [[r\`]]
    x[115] = [[s]]
    x[642] = [[s`]]
    x[597] = [[s\]]
    x[116] = [[t]]
    x[648] = [[t`]]
    x[117] = [[u]]
    x[118] = [[v]]
    x[651] = [[v\]]
    x[119] = [[w]]
    x[120] = [[x]]
    x[615] = [[x\]]
    x[121] = [[y]]
    x[122] = [[z]]
    x[656] = [[z`]]
    x[657] = [[z\]]
    x[593] = [[A]]
    x[946] = [[B]]
    x[665] = [[B\]]
    x[231] = [[C]]
    x[240] = [[D]]
    x[603] = [[E]]
    x[625] = [[F]]
    x[611] = [[G]]
    x[610] = [[G\]]
    x[667] = [[G\_<]]
    x[613] = [[H]]
    x[668] = [[H\]]
    x[618] = [[I]]
    x[407] = [[I\]]
    x[626] = [[J]]
    x[607] = [[J\]]
    x[644] = [[J\_<]]
    x[620] = [[K]]
    x[622] = [[K\]]
    x[654] = [[L]]
    x[671] = [[L\]]
    x[623] = [[M]]
    x[624] = [[M\]]
    x[331] = [[N]]
    x[628] = [[N\]]
    x[596] = [[O]]
    x[664] = [[O\]]
    x[594] = [[Q]]
    x[641] = [[R]]
    x[640] = [[R\]]
    x[643] = [[S]]
    x[952] = [[T]]
    x[650] = [[U]]
    x[650] = [[U\]]
    x[652] = [[V]]
    x[653] = [[W]]
    x[967] = [[X]]
    x[295] = [[X\]]
    x[655] = [[Y]]
    x[658] = [[Z]]
    x[46] = [[.]]
    x[712] = [["]]
    x[716] = [[%]]
    x[690] = [[_j]]
    x[720] = [[:]]
    x[721] = [[:\]]
    x[601] = [[@]]
    x[600] = [[@\]]
    x[230] = [[{]]
    x[649] = [[}]]
    x[616] = [[1]]
    x[248] = [[2]]
    x[604] = [[3]]
    x[606] = [[3\]]
    x[638] = [[4]]
    x[619] = [[5]]
    x[592] = [[6]]
    x[612] = [[7]]
    x[629] = [[8]]
    x[339] = [[9]]
    x[630] = [[&]]
    x[660] = [[?]]
    x[661] = [[?\]]
    x[674] = [[<\]]
    x[673] = [[>\]]
    x[8593] = [[^]]
    x[8595] = [[!]]
    x[451] = [[!\]]
    x[124] = [[|]]
    x[448] = [[|\]]
    x[8214] = [[||]]
    x[449] = [[|\|\]]
    x[450] = [[=\]]
    x[814] = [[-\]]
    x[776] = [[_"]]
    x[799] = [[_+]]
    x[800] = [[_-]]
    x[711] = [[_/]]
    x[805] = [[_0]]
    x[809] = [[_=]]
    x[700] = [[_>]]
    x[740] = [[_?\]]
    x[710] = [[_\]]
    x[815] = [[_^]]
    x[794] = [[_}]]
    x[734] = [[`]]
    x[771] = [[_~]]
    x[792] = [[_A]]
    x[826] = [[_a]]
    x[783] = [[_B]]
    x[796] = [[_c]]
    x[810] = [[_d]]
    x[820] = [[_e]]
    x[770] = [[_F]]
    x[736] = [[_G]]
    x[833] = [[_H]]
    x[688] = [[_h]]
    x[816] = [[_k]]
    x[832] = [[_L]]
    x[737] = [[_l]]
    x[772] = [[_M]]
    x[827] = [[_m]]
    x[828] = [[_N]]
    x[8319] = [[_n]]
    x[825] = [[_O]]
    x[798] = [[_o]]
    x[793] = [[_q]]
    x[780] = [[_R]]
    x[797] = [[_r]]
    x[779] = [[_T]]
    x[804] = [[_t]]
    x[812] = [[_v]]
    x[695] = [[_w]]
    x[774] = [[_X]]
    x[829] = [[_x]]    
    -- 
    for p, c in utf8.codes(s) do
        b = b..x[c]
    end
    return b
end


function WLiterary(i)
  WGoodText(i,0)
end


function WNotes(i)
  WGoodText(i,1)
end


function WSort(i, st)
  if st == nil then st = stNormal end
  WGoodText(i,2,st)
end


-- Counting the frequency of words in the text
function WordFrequency(s, g)
    if g == nil then g = true end
    os.setlocale('', 'ctype') -- to support regional locale
    local function delzer(s)
        local c
        local a = string.split(s,'|')
        if #a == 2 then
            c = string.format("%u",a[1])..'|'..a[2]
        else
            c = s
        end
        return c
    end
    w = string.split(s, ' \r\n\9%-')
    -- create associative frequency table
    if g then GaugeInit(#w) end
    local a = {}
    for i = 1,#w do
        b = string.lower(Body(w[i]))
        if string.trim(b) ~= '' then
            if a[b] == nil then
                a[b] = 1
            else
                a[b] = a[b]+1
            end
        end
        if g and i % 1024 == 0 then Gauge(i) end
    end
    if g then Gauge(0) end
    -- sort by decreasing frequency
    q = {}
    i = 0
    for key, val in pairs(a) do
        c = string.format("%06d",val)
        i = i+1
        q[i] = c..'|'..key
    end
    table.sort(q, function(x,y) return (x > y) end)
    -- delete leading zeros in numbers
    for i = 1,#q do
        q[i] = delzer(q[i])
    end 
    os.setlocale('C')  -- default locale
    return q
end


-- Cancel script when user pressed global hotkey Break
-- Thank tonio_k
function CancelScript()
    local msg, umsg
    if CurrentLang() == 'Russian' then
        msg = '  !'
        umsg = AnsiToUtf8(msg,true)
    else
        msg = 'SCRIPT TERMINATED BY USER!'
        umsg = msg
    end
    if Cancel() then
        Gauge(0) -- remove progress bar (if was activated) 
        StatusMessage(msg)
        error(umsg..'\r\r',0) 
    end
end


-- Utility functions for the option "Find / Replace" by Lua pattern --
--
-- Get fields: Find-FindText; Repl-FindText; Repl-ReplText
-- and the number of active tab: '0' - Find; '1' - Replace
function GetSearchPanelFields()
    local c = OtherUtils('',4)
    local t = StrSplit(c,'\3')
    return t[1], t[2], t[3], t[4], t[5], t[6]
end

-- Find from current cursor position and highlight
function frFindNext()
    os.setlocale('', 'ctype')
    local k = WActive()
    local start, len
    if WSelText(k) == '' then
        start = WSel(k)
    else
        start, len = WSel(k)
        start = start+len+1
    end
    local s = WText(k)
    local eFR1, eFR2, _, aPage = GetSearchPanelFields()
    local i, j, e, f, g
    -- may be search pattern or @... the small script returning the search pattern!
    if aPage == '0' then
        if string.sub(eFR1,1,1) == '@' then f = load(string.sub(eFR1,2,#eFR1)); if f then e = f(); if e then eFR1 = e end end end
        f = load('p =\34'..eFR1..'\34'); f()
    end
    if aPage == '1' then
        if string.sub(eFR2,1,1) == '@' then g = load(string.sub(eFR2,2,#eFR2)); if g then e = g(); if e then eFR2 = e end end end
        g = load('p =\34'..eFR2..'\34'); g()
    end
    i, j = string.find(s,p,start)  -- search
    if i then  -- highlight found
        WSel(k,i-1,j-i+1)
    else
        WSel(k,1) -- go to top of text
    end
    os.setlocale('C')
end

-- Replace highlighted
function frReplace()
    os.setlocale('', 'ctype')
    local k = WActive()
    local b = WSelText(k)
    local _, eFR2, eFR3, aPage = GetSearchPanelFields()
    if aPage == '0' then return end
    if aPage == '1' then
        f = load('p =\34'..eFR2..'\34'); f()
        g = load('q =\34'..eFR3..'\34'); g()
    end
    b = string.gsub(b,p,q)
    WSelText(k,b)
    os.setlocale('C')
end

-- Calculate all suitable fragments
function frNumberOfSuitable()
    os.setlocale('', 'ctype')
    local eFR1, _, _, aPage, msg = GetSearchPanelFields()
    local c, n, e, f
    local k = WActive()
    local s = WText(k)
    if aPage == '1' then return end
    if string.sub(eFR1,1,1) == '@' then f = load(string.sub(eFR1,2,#eFR1)); if f then e = f(); if e then eFR1 = e end end end
    f = load('p =\34'..eFR1..'\34'); f()
    n = 0
    start = 1
    i = 1
    while i do
        i, j = string.find(s,p,start)
        if i then
            n = n+1
            start = j+1
        end
    end
    f = SplitFileName(WName(k))
    msg = string.gsub('# '..msg,'%%s',f[2]..f[3])
    WNew(0,msg..'\r\r')
    WAdd(0,eFR1..' - '..n)
    WActive(0)
    os.setlocale('C')
end

-- List of all suitable fragments from text
function frAllSuitable()
    os.setlocale('', 'ctype')
    local eFR1, _, _, aPage, _, msg = GetSearchPanelFields()
    local i, j, start, b, c, e, f
    local k = WActive()
    local s = WText(k)
    if aPage == '1' then return end
    if string.sub(eFR1,1,1) == '@' then f = load(string.sub(eFR1,2,#eFR1)); if f then e = f(); if e then eFR1 = e end end end
    f = load('p =\34'..eFR1..'\34'); f()
    b = {}
    start = 1
    i = 1
    while i do
        i, j = string.find(s,p,start)
        if i then
            c = string.sub(s,i,j)
            b[#b+1] = c
            start = j+1
        end
    end
    c = table.concat(b,'\r')
    f = SplitFileName(WName(k))
    msg = string.gsub('# '..msg,'%%s',f[2]..f[3])
    WNew(0,msg..'\r\r')
    WAdd(0,c)
    WActive(0)
    os.setlocale('C')
end

-- Replace all in plain text
function frTotalReplace()
    os.setlocale('', 'ctype')
    local k = WActive()
    local s = WText(k)
    local _, eFR2, eFR3, aPage = GetSearchPanelFields()
    local start = WSel(k)
    if aPage == '0' then return end
    if aPage == '1' then
        f = load('p =\34'..eFR2..'\34'); f()
        g = load('q =\34'..eFR3..'\34'); g()
    end
    s = string.gsub(s,p,q)
    WNew(k,s,false)
    WSel(k,start)
    os.setlocale('C')
end

-- Replace all while maintaining formatting rtf/dxt (works slowly)
function frFancyTotalReplace()
    os.setlocale('', 'ctype')
    local _, eFR2, eFR3, aPage = GetSearchPanelFields()
    local i, j, start, start0, b
    local k = WActive()
    local s = WText(k)
    if aPage == '0' then return end
    if aPage == '1' then
        f = load('p =\34'..eFR2..'\34'); f()
        g = load('q =\34'..eFR3..'\34'); g()
    end
    start0 = WSel(k); if start0 == 0 then start0 = 1 end
    start = 1
    i = 1
    GaugeInit(#s)
    while i do
        i, j = string.find(s,p,start)
        if i then
            WSel(k,i-1,j-i+1)
            b = string.sub(s,i,j)
            b = string.gsub(b,p,q)
            WSelText(k,b)
            s = WText(k)
            start = i+#b
            if i % 1024 == 0 then Gauge(i) end
        end
    end
    Gauge(0)
    WSel(k,start0)
    os.setlocale('C')
end
-- End of utility functions for the option "Find / Replace" by Lua pattern --


-- Converter of dic rules to Lua patterns (beta version)
function DicToLua(d)
    local b = '%w'
    local x = '%s+'
    local y = '%s+'
    -- -- preparation, solution of simple cases -- --
    d = string.trim(d)
    if d == '' then return '' end  -- -- -->>>
    if not string.find(d,'=',1,true) or (string.sub(d,1,1) == '=') then return '' end
    if (string.sub(d,1,1) == '#') then  -- maybe: comment sign
        if (string.sub(d,2,2) ~= '#') then return '' else d = string.sub(d,2,#d) end  -- -- -->>>     
    end
    if (string.sub(d,1,1) == '$') then  -- maybe: case-sensitive sign
        if (string.sub(d,2,2) ~= '$') then d = '@'..string.sub(d,2,#d) else d = string.sub(d,2,#d) end    
    end
    local p = string.split(d,'=')
    if p[2] == nil then p[2] = ' ' end
    if string.trim(string.gsub(p[1],'%*','')) == '' then  -- left part of one * - to shield everything!
        p[1] = string.gsub(p[1],'%*','%%*') 
        return p[1]..'='..p[2]  -- -- -->>>
    end
    local c = ''
    local fnd = false
    for j = 1,#p[1] do  -- shield characters matching magic except *
        x = string.sub(p[1],j,j) 
        if string.find('().%+-?[]^$',x,1,true) then fnd = true; c = c..'%'..x else c = c..x end
    end
    if fnd then p[1] = c end
    -- -- analysis of the rule in the general case -- --
    local ws, wd = string.split(p[1],' '), string.split(p[2],' ')
    if wd == nil then wd = {''} end
    local m = #ws
    local m1 = #wd
    local z = '('..b..'-)'  -- an asterisk at the beginning / end of the dic pattern
    local u = '('..b..'+)'  -- freestanding asterisk
    local k = 0 
    for i = 1, m do
        if i > m1 then table.insert(wd,'') end 
        if ws[i] ~='*' then
            if string.sub(ws[i],1,1) ~= '*' and string.sub(ws[i],#ws[i],#ws[i]) ~= '*' then
                ws[i] = '%f[%w]'..ws[i]..'%f[%W]'
            else
            if string.sub(ws[i],1,1) == '*' then 
                k = k+1
                ws[i] = z..string.sub(ws[i],2,#ws[i])
                if string.sub(ws[i],#ws[i],#ws[i]) ~= '*' then ws[i] = ws[i]..'%f[%W]' end
                wd[i] = '%'..tostring(k)..wd[i]
            end
            if string.sub(ws[i],#ws[i],#ws[i]) == '*' then
                k = k+1
                ws[i] = string.sub(ws[i],1,#ws[i]-1)..z
                if string.sub(ws[i],1,1) ~= '*' then ws[i] = '%f[%w]'..ws[i] end
                wd[i] = wd[i]..'%'..tostring(k)
            end
            end
        end
        if ((ws[i] == '*') or (ws[i] == '$*')) and (wd[i] ~= '') then
            k = k+1
            ws[i] = u
            table.insert(wd,i,'%'..tostring(k))
        end
    end
    return table.concat(ws,' ')..'='..table.concat(wd,' ')  -- -- -->>>
end


-- Analogue for Lua patterns of functions RexMatch and DicMatch
function LuaMatch(s, p, k)
    if k == nil then k = 1 end
    local fnd, pos = '', 0
    os.setlocale('', 'ctype')
    i, j = string.find(s,p,k)
    if i then
        fnd = string.sub(s,i,j)
        pos = i
    end
    os.setlocale('C')
    return fnd, pos
end


-- Analogue for Lua patterns of functions RexRepl and DicRepl
-- s - source text
-- r - table of lines with Lua-rules: {'search1=replace1', 'search2=replace2', ... }
-- example - deleting the double spaces: {'[ ]+= '} 
-- instead of = you can use ::
function LuaRepl(s, r, gauge)
    local function LuaOneRuleRepl(s,p)
        os.setlocale('', 'ctype')
        local k, left, right
        k = string.find(p,'::',1,true)
        if k then
            left = string.sub(p,1,k-1)
            right = string.sub(p,k+2,#p)
        else
            k = string.find(p,'=',1,true)
            left = string.sub(p,1,k-1)
            right = string.sub(p,k+1,#p)
        end
        -- ! now replace in s !
        if k then
            c = 'u = string.gsub(u,\34'..left..'\34,\34'..right..'\34)'
            u = s
            f = load(c); f() 
        end
        s = u
        os.setlocale('C')
        return s
    end
    if gauge == nil then gauge = true end
    if gauge then GaugeInit(#r) end
    for i = 1,#r do
        s = LuaOneRuleRepl(s,r[i])
        if gauge then Gauge(i) end
    end
    if gauge then Gauge(0) end
    return s
end


-- Find all matches for Lua / Rex / Dic pattern
function MatchAll(text, pattern, mode, gauge)
    if gauge == nil then gauge = true end
    local s = text..'\r'
    local r = pattern
    local k = 1
    local temp = {}
    local xmatch
    local fnd
    local u = 0
    if gauge then GaugeInit(#s) end
    if mode == maLua then
        xmatch = LuaMatch
    elseif mode == maRex then
        xmatch = RexMatch
    elseif mode == maDic then
        xmatch = DicMatch
    else
        ShowMessage('MatchAll: 3-rd parameter - wrong pattern type.\rMust be: maLua or maRex or maDic.')
        return {}, 0
    end       
    while k > 0 do
        CancelScript()
        fnd, k = xmatch(s,r,k)
        if k > 0 then
            if gauge then Gauge(k) end
            u = u+1
            temp[u] = fnd
            k = k+#fnd 
        else
            break
        end
    end
    if gauge then Gauge(0) end   
    return temp, u
end


-- Reduction paragraphs in the text to a size not exceeding the specified limit --
-- Paragraphs that are too large are divided into smaller paragraphs - by the signs
-- of the end of sentences, or punctuation marks, or in extreme cases - by spaces
-- between words
-- Based on the initial development by tonio_k

-- name of the hotkey assigned in Settings to interrupt long-running processes;
-- if not assigned, nil will return
function BreakKeyName()
    local bk
    local a = LoadFromFile(HomeFolder()..'$.cfg',true)
    for i = 1,#a do
        if #a[i] > 8 and string.sub(a[i],1,8) == 'GKName4=' then
            bk = string.sub(a[i],9,#a[i])
            bk = string.trim(bk)
            if bk ~= '' then return bk else return nil end
        end
    end
    return nil
end


-- remove "garbage" in the text
function GoodText(s)
    s = string.gsub(s,'[ ]+',' ') -- remove double spaces
    s = string.gsub(s,'^( +)','') -- remove spaces at the beginning of the text
    s = string.gsub(s, '^[\r\n]+','') -- remove a blank paragraphs at the beginning of the text
    s = string.gsub(s,'[\r\n][\32]+','\r') -- remove spaces at the beginning of the each paragraph
    s = string.gsub(s,'[\r\n][\32]+[\r\n]','\r\r') -- remove spaces in...
    s = string.gsub(s,'[\r\n]+\32[\r\n]+','\r\r') -- ... empty lines
    s = string.gsub(s,'[\r\n]+','\r') -- remove empty lines
    return s
end


-- finds a paragraph in the text that exceeds the specified limit
-- if there are several, then returns the first one found
function LargeParag(s, k, limit)
    local b, p
    s = s..'\r'
    p = k
    i = k
    while i < #s do
       b = string.sub(s,i,i)
       if b == '\r' then
           if i-p+1 > limit then 
               return string.sub(s,p,i), p
           else
               i = i+1
               p = i
           end
       end
       i = i+1
    end
    return '', 0
end


-- largest initial substring s, not longer than d
-- and ending with one of the characters of the string v
-- example: StrStub(s,500,'.!?')
function StrStub(s, d, v)
    local i, j, b, c
    j = 0
    i = 1
    while i < #s do
        b = string.sub(s,i,i)
        c = string.sub(s,i+1,i+1)
        if string.find(v,b,1,true) and not string.find(v,c,1,true) and i < d then
            j = i
        end
        i = i+1
        if i > d then break end
    end
    if j == 0 then return '' else return string.sub(s,1,j) end
end


-- breaks a large paragraph into smaller ones
function DivLargeParag(abzac, limit)
    local k0 = 1
    local k_temp0 = 1
    local fnd0 = ''
    local d = math.floor(0.618*#abzac)
    if CurrentLang() == 'Russian' then
        umsg = AnsiToUtf8('  !\13    ?',true)
    else
        umsg = 'No spaces found!\13Specified paragraph size is too small?'
    end
    -- 1) try to break the paragraph in half according to the signs of the end of the sentence [.?!]
    fnd0 = StrStub(abzac,d,'.?!')
    if #fnd0 > 0 then
        abzac = fnd0.."\r"..string.sub(abzac,#fnd0+1,#abzac)
    end
    -- 2) try to break the paragraph by the end of the sentence [.?!]
    if #fnd0 == 0 then
        fnd0 = StrStub(abzac,limit,'.?!')
        if #fnd0 > 0 then
            abzac = fnd0..'\r'..string.sub(abzac,#fnd0+1,#abzac)
        end
    end
    -- 3) try to break a paragraph by punctuation [,:;/\]
    if #fnd0 == 0 then
        fnd0 = StrStub(abzac,limit,[[,:;/\]])
        if #fnd0 > 0 then
            abzac = fnd0..'\r'..string.sub(abzac,#fnd0+1,#abzac)
        end
    end
    -- 4)      []
    if #fnd0 == 0 then
        fnd0 = StrStub(abzac,limit,' ')
        if #fnd0 == 0 then
            error(umsg..'\r\r',0)
        end
        if #fnd0 > 0 then
            abzac = fnd0..'\r'..string.sub(abzac,#fnd0+1,#abzac)
        end
    end
    return abzac -- splitted paragraphend
end


-- limit the size of paragraphs in the text
-- by dividing large into smaller ones
function ParagsLimitation(text, limit, gauge)
    local s, k, fnd, s_start, s_end = text..'\r', 1, '', 1, 1
    s = GoodText(s)
    if gauge == nil then gauge = true end
    if gauge then GaugeInit(#s) end
    while k < #s do
        fnd, k = LargeParag(s,k,limit)
        if #fnd > 0 then
            s_start = string.sub(s,1,k-1)
            s_end = string.sub(s,k+#fnd,#s)
            fnd = DivLargeParag(fnd,limit)
            s = s_start..fnd..s_end -- update the working string
        else
            break
        end
        if gauge then Gauge(k) end
        CancelScript() -- user interrupt check
    end
    if gauge then Gauge(0) end
    return s
end
-- End of Reduction paragraphs ... --
